package org.nakedobjects.nos.remote.command.socket;

import java.io.IOException;
import java.net.SocketException;

import org.apache.log4j.Logger;
import org.nakedobjects.noa.adapter.NakedObjectLoader;
import org.nakedobjects.noa.security.Session;
import org.nakedobjects.noa.security.ConnectionManager;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.core.system.Monitor;
import org.nakedobjects.nof.core.util.DebugString;
import org.nakedobjects.nos.remote.command.Request;
import org.nakedobjects.nos.remote.command.Response;


public class Worker implements Runnable {
    private static final Logger LOG = Logger.getLogger(Worker.class);
    private static final Logger ACCESS_LOG = Logger.getLogger("access_log");
    private static int nextId = 1;
    private ServerConnection connection;
    private final ConnectionManager connectionManager;
    private final WorkerPool poolToReturnTo;
    private final int id = nextId++;
    private String requestDebug;
    private String sessionDebug;
    private String responseDebug;
    private String contextId;
    private long responseTime;
    private boolean running = true;

    public Worker(final ConnectionManager connectionManager, WorkerPool pool) {
        this.connectionManager = connectionManager;
        this.poolToReturnTo = pool;
    }

    public synchronized void gracefulStop() {
        running = false;
    }

    public synchronized void setConnection(final ServerConnection connection) {
        this.connection = connection;
        notify();
    }

    public boolean isAvailable() {
        return connection == null;
    }

    public synchronized void run() {
        running: while (running) {
            while (connection == null) {
                try {
                    wait();
                } catch (InterruptedException e) {
                    if (!running) {
                        LOG.info("Request to stop : " + toString());
                        break running;
                    }
                }
            }

            try {
                // TODO client should request to disconnect. then session should be closed
                long start = System.currentTimeMillis();
                Request request = connection.awaitRequest();
                Session session = null;
                try {
                    requestDebug = request.getId() + " - " + request.toString();
                    contextId = NakedObjectsContext.getExecutionContextId();
                    session = request.getSession();
                    sessionDebug = session == null ? "none" : session.toString();
                    connectionManager.connect(session);
                    request.execute(connection.getServer());
                    Response response;
                    response = new Response(request);
                    responseDebug = response.toString();
                    LOG.debug("sending " + response);
                    connection.sendResponse(response);

                    String message = "request " + request.toString();
                    ACCESS_LOG.info(message);
                    Monitor.addEvent("Server", message);
                } catch (Exception e) {
                    LOG.error("error during remote request", e);
                    connection.sendResponse(e);
                }
                if (session != null) {
                    connectionManager.disconnect(session);
                }
                responseTime = System.currentTimeMillis() - start;
                end();
            } catch (SocketException e) {
                LOG.info("shutting down receiver (" + e + ")");
                end();
            } catch (IOException e) {
                LOG.info("connection exception; closing connection", e);
                end();
            }
        }
        LOG.info("Stopping: " + toString());
    }

    private void end() {
        connection = null;
        poolToReturnTo.returnWorker(this);
    }

    public void debug(DebugString debug) {
        debug.appendln("context", contextId);
        debug.appendln("session", sessionDebug);
        debug.appendln("request", requestDebug);
        debug.appendln("response", responseDebug);
        debug.appendln("duration", responseTime / 1000.0f + " secs.");

        NakedObjectLoader loader = NakedObjectsContext.getObjectLoader();
        if (loader != null) {
            // ((DebugInfo) loader).debugData(debug);
        }
    }

    public String toString() {
        return "Worker#" + id;
    }
}

// Copyright (c) Naked Objects Group Ltd.
